from typing import Optional

import torch
from torch.autograd import Function


__all__ = ["absmax_scale", "dtype_info", "QTensor"]


def dtype_info(dtype):
    info = torch.finfo if dtype.is_floating_point else torch.iinfo
    return info(dtype)


def absmax_scale(
    base: torch.Tensor, itype: torch.Tensor.dtype = torch.int8, axis: Optional[int] = None
) -> torch.Tensor:
    """Evaluate the quantization scale using the absmax algorithm.

    The Absolute Maximum quantization algorithm is a symmetrical quantization
    algorithm where the scale corresponds to the maximum absolute value of the
    base divided by the highest positive integer value for the target integer
    representation.

    Args:
        base (`torch.Tensor`): the base tensor on which the scale will be applied.
        itype (`torch.Tensor.dtype`): the target internal dtype for quantization.
        axis (`int`): the index of the axis to preserve, or -1 for the last one.
            Defaults to None to reduce all axis.

    Returns:
        `torch.Tensor`: a scale tensor of the same dtype as the base.
    """
    abs_base = torch.abs(base)
    if axis is None:
        qranges = torch.max(abs_base)
    else:
        dim = list(range(base.ndim))
        if axis == -1:
            dim = dim[:-1]
        else:
            dim.remove(axis)
        qranges = torch.amax(torch.abs(base), dim=dim, keepdim=True)
    info = dtype_info(itype)
    return qranges / info.max


class Quantizer(Function):
    """A standard symmetric quantizer.

    If the quantization scale is not specified, it uses the absmax scale for the base tensor.
    """

    @staticmethod
    def forward(ctx, base, itype: torch.Tensor.dtype = torch.int8, scale=None):
        info = dtype_info(itype)
        if scale is None:
            scale = absmax_scale(base, itype)
        elif scale.ndim > 0:
            if torch.squeeze(scale).ndim > 1:
                raise ValueError("Quantizing along multiple axis is not supported")
            if scale.ndim != base.ndim:
                raise ValueError(
                    "When quantizing per-axis, the scale must be broadcastable to the base (Tip: try to add missing dims of length zero)."
                )
        data = base / scale
        if not itype.is_floating_point:
            data = torch.round(data)
        data = torch.clamp(data, min=info.min, max=info.max).to(itype)
        # The instantiation of the quantized tensor must happen within the context of the Function
        # for the autograd magic to work.
        return QTensor(data, scale)

    @staticmethod
    def backward(ctx, gO):
        # For autograd, quantization is a no-op
        return gO, None, None


class Dequantizer(Function):
    @staticmethod
    def forward(ctx, t):
        if t.itype == torch.int32:
            # The dequantization operation requires data to be cast to the scale float type before multiplication
            # by the scale, but this might actually overflow for float16/bfloat16
            return (t._scale.to(torch.float32) * t._data).to(t._scale.dtype)
        elif t.itype.is_floating_point:
            # Upcast explicitly to the scale dtype
            return t._scale * t._data.to(t._scale.dtype)
        return t._scale * t._data

    @staticmethod
    def backward(ctx, gO):
        # For autograd, dequantization is a no-op
        return gO


class ReQuantizer(Function):
    @staticmethod
    def forward(ctx, base, itype=torch.int8, scale=None):
        dst_info = dtype_info(itype)
        if scale is None:
            if itype == base.itype:
                return base
            # Assuming the base scale is correct, simply project to the target integer range
            src_info = dtype_info(base.itype)
            int_rescale = dst_info.max / src_info.max
            scale = base._scale * src_info.max / dst_info.max
        else:
            # It is up to the caller to make sure the scale is consistent with the target int dtype
            int_rescale = base._scale / scale
        if base.itype == torch.int32:
            # The rescaling operation requires data to be cast to the scale float type before multiplication
            # by the scale, but this might actually overflow for float16/bfloat16
            int_rescale = int_rescale.to(torch.float32)
        data = torch.clamp(torch.round(base._data * int_rescale), min=dst_info.min, max=dst_info.max).to(itype)
        # The instantiation of the quantized tensor must happen within the context of the Function
        # for the autograd magic to work.
        return QTensor(data, scale)

    @staticmethod
    def backward(ctx, gO):
        # For autograd, requantization is a no-op
        return gO, None, None


class QTensor(torch.Tensor):
    @staticmethod
    def __new__(cls, data, scale, requires_grad=False):
        assert data.device == scale.device
        # This constructor can ONLY create leaf Tensors wrt autograd.
        # Use QTensor.from_tensor(t) to get a non-leaf Tensor wrt autograd.
        return torch.Tensor._make_wrapper_subclass(
            cls, data.size(), strides=data.stride(), dtype=scale.dtype, device=data.device, requires_grad=requires_grad
        )

    def __init__(self, data, scale, requires_grad=False):
        self._axis = None
        if scale.ndim > 0:
            if torch.squeeze(scale).ndim > 1:
                raise ValueError("QTensor cannot be quantized along multiple axis.")
            if scale.ndim != data.ndim:
                raise ValueError(
                    "The QTensor scale must be broadcastable to the base (Tip: try to add missing dims of length zero)."
                )
            dims = scale.shape
            for i in range(scale.ndim):
                if dims[i] != 1:
                    self._axis = i
            if self._axis is None:
                # All dims are 1: the scale is actually a scalar
                scale = torch.squeeze(scale)
        self._data = data
        self._scale = scale

    def __repr__(self):  # Zero out missing values for printing
        autograd_info = (
            f", grad_fn={self.grad_fn}" if self.grad_fn else ", requires_grad=True" if self.requires_grad else ""
        )
        return f"QTensor({self._data}, scale={self._scale}, public_dtype={self.dtype}{autograd_info})"

    @classmethod
    def quantize(cls, base, itype=torch.int8, scale=None):
        """Differentiable quantization function"""
        return Quantizer.apply(base, itype, scale)

    def dequantize(self):
        """Differentiable dequantization function"""
        return Dequantizer.apply(self)

    def rescale(self, itype=torch.int8, scale=None):
        """Differentiable requantization function"""
        return ReQuantizer.apply(self, itype, scale)

    @property
    def axis(self):
        return self._axis

    @property
    def itype(self):
        return self._data.dtype

    def __tensor_flatten__(self):
        return ["_data", "_scale"], None

    @staticmethod
    def __tensor_unflatten__(inner_tensors, meta, outer_size, outer_stride):
        assert len(inner_tensors) == 2
        assert meta is None
        data, scale = inner_tensors["_data"], inner_tensors["_scale"]
        return QTensor(data, scale)

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        from .func import get_qtensor_func

        kwargs = kwargs or {}

        # Look for a func accepting QTensor inputs
        qfunc = get_qtensor_func(func)
        if qfunc is not None:
            return qfunc(*args, **kwargs)
        # Defer to dispatcher to look instead for QTensor operations
        with torch._C.DisableTorchFunctionSubclass():
            return func(*args, **kwargs)

    @classmethod
    def __torch_dispatch__(cls, op, types, args, kwargs=None):
        from .ops import get_qtensor_op

        # Do not use directly func, but rather its overload
        op = op.overloadpacket
        # Look for a dispatched op accepting QTensor inputs
        qop = get_qtensor_op(op)
        if qop is not None:
            return qop(*args, **kwargs)
        # Identify the types of the args
        types = [type(arg).__name__ for arg in args]
        raise ValueError(f"{op} {types} is no supported for QTensor.")

    def numpy(self):
        return self.dequantize().cpu().numpy()



import io

import pytest
import torch
from helpers import random_qtensor, random_tensor

from quanto.quantization import QTensor, absmax_scale, calibration, freeze, quantize
from quanto.quantization.nn import QLinear, QModuleMixin


@pytest.mark.parametrize("input_shape", [(10,), (1, 10), (2, 10), (10, 32, 32)])
@pytest.mark.parametrize("itype", [torch.int8], ids=["int8"])
@pytest.mark.parametrize("dtype", [torch.float16, torch.float32], ids=["fp16", "fp32"])
@pytest.mark.parametrize("axis", [None, 0, -1], ids=["per-tensor", "first-axis", "last-axis"])
def test_quantized_tensor_serialization(input_shape, itype, dtype, axis):
    inputs = random_tensor(input_shape, dtype=dtype)
    scale = absmax_scale(inputs, itype, axis)
    qinputs = QTensor.quantize(inputs, itype, scale)
    b = io.BytesIO()
    torch.save(qinputs, b)
    b.seek(0)
    qinputs_reloaded = torch.load(b)
    assert torch.equal(qinputs_reloaded._data, qinputs._data)
    assert torch.equal(qinputs_reloaded._scale, qinputs._scale)
    # We cannot test dtype directly, as it is not set correctly by torch.load
    assert qinputs_reloaded._scale.dtype == dtype
    assert qinputs_reloaded.axis == qinputs.axis


@pytest.mark.parametrize("use_bias", [True, False], ids=["bias", "no-bias"])
@pytest.mark.parametrize("weights", [torch.int8], ids=["w-int8"])
@pytest.mark.parametrize("activations", [None, torch.int8], ids=["a-float", "a-int8"])
@pytest.mark.parametrize("dtype", [torch.float16, torch.float32], ids=["fp16", "fp32"])
def test_qlinear_serialization(use_bias, activations, weights, dtype, device):
    if dtype == torch.float16 and device.type == "cpu":
        pytest.skip("Matrix multiplication is not supported for float16 on CPU")
    embeddings = 10
    linear = torch.nn.Linear(embeddings, embeddings, bias=use_bias).to(dtype).to(device)
    qlinear = QLinear.from_module(linear, weights=weights, activations=activations)
    if activations is not None:
        qinputs = random_qtensor((10, 10, embeddings), dtype=dtype).to(device)
        with calibration():
            qlinear(qinputs)
    qlinear.freeze()
    b = io.BytesIO()
    torch.save(qlinear.state_dict(), b)
    b.seek(0)
    state_dict = torch.load(b)
    qlinear_reloaded = QLinear(embeddings, embeddings, bias=use_bias)
    # We need to force assignment instead of copy to replace weights by quantized weights
    qlinear_reloaded.load_state_dict(state_dict, assign=True)
    w = qlinear.weight
    w_reloaded = qlinear_reloaded.weight
    assert torch.equal(w._data, w_reloaded._data)
    assert torch.equal(w._scale, w_reloaded._scale)
    assert w_reloaded.dtype == dtype
    assert w_reloaded.axis == w.axis
    if activations is not None:
        for attr in ["input_scale", "output_scale"]:
            v = getattr(qlinear, attr)
            v_reloaded = getattr(qlinear_reloaded, attr)
            assert torch.equal(v, v_reloaded)


class MLP(torch.nn.Module):
    def __init__(self, input_size, output_size, hidden_size):
        super().__init__()
        self.input_layer = torch.nn.Linear(input_size, hidden_size)
        self.mid_layer = torch.nn.Linear(hidden_size, hidden_size)
        self.output_layer = torch.nn.Linear(hidden_size, output_size)

    def forward(self, inputs):
        x = torch.nn.functional.relu(self.input_layer(inputs))
        x = torch.nn.functional.relu(self.mid_layer(x))
        return torch.nn.functional.softmax(self.output_layer(x), dim=-1)


@pytest.mark.parametrize("weights", [torch.int8], ids=["w-int8"])
@pytest.mark.parametrize("dtype", [torch.float16, torch.float32], ids=["fp16", "fp32"])
def test_serialize_quantized_mlp(weights, dtype, device):
    if dtype == torch.float16 and device.type == "cpu":
        pytest.skip("Matrix multiplication is not supported for float16 on CPU")
    input_features = 32
    hidden_features = 10
    output_features = 128
    model = MLP(input_features, hidden_features, output_features).to(dtype).to(device)
    quantize(model, weights=weights)
    inputs = random_tensor((1, 10, input_features), dtype=dtype).to(device)
    with calibration():
        model(inputs)
    freeze(model)
    b = io.BytesIO()
    torch.save(model.state_dict(), b)
    b.seek(0)
    state_dict = torch.load(b)
    model_reloaded = MLP(input_features, hidden_features, output_features).to(device)
    quantize(model_reloaded)
    # When reloading we must assign instead of copying to force quantized tensors assignment
    model_reloaded.load_state_dict(state_dict, assign=True)
    for name, module in model.named_modules():
        if isinstance(module, QModuleMixin):
            module_reloaded = getattr(model_reloaded, name)
            assert torch.equal(module_reloaded.weight._data, module.weight._data)
            assert torch.equal(module_reloaded.weight._scale, module.weight._scale)
            assert torch.equal(module_reloaded.input_scale, module.input_scale)
            assert torch.equal(module_reloaded.output_scale, module.output_scale)



from typing import Optional

import torch

from ..qtensor import QTensor, absmax_scale
from .qmodule import QModuleMixin, register_qmodule


__all__ = ["QLinear"]


@register_qmodule(torch.nn.Linear)
class QLinear(QModuleMixin, torch.nn.Linear):
    def __init__(self, *args, weights: torch.dtype = torch.int8, **kwargs):
        super().__init__(*args, **kwargs)
        self.weights = weights

    @classmethod
    def from_module(cls, module, weights=torch.int8, activations: Optional[torch.dtype] = None):
        qmodule = cls(
            module.in_features,
            module.out_features,
            module.bias is not None,
            dtype=module.weight.dtype,
            weights=weights,
            activations=activations,
        )
        with torch.no_grad():
            qmodule.weight.copy_(module.weight)
            if module.bias is not None:
                qmodule.bias.copy_(module.bias)
        return qmodule.to(module.weight.device)

    def qweight(self):
        if isinstance(self.weight, QTensor):
            return self.weight
        # Quantize the weights per-axis
        wscale = absmax_scale(self.weight, axis=0)
        return QTensor.quantize(self.weight, itype=self.weights, scale=wscale)

    def qforward(self, input: torch.Tensor) -> torch.Tensor:
        if self.activations is not None and not isinstance(input, QTensor):
            # Quantize tensor to be able to take advantage of accelerated matmul
            input = QTensor.quantize(input, self.activations, self.input_scale)
        # We always use quantized weights
        qweight = self.qweight()
        output = torch.matmul(input, qweight.t())
        if self.bias is not None:
            # The outputs will be dequantized in the addition since the biases are not quantized
            output = output + self.bias
        return output



from typing import Optional

import torch

from .qmodule import QModuleMixin, register_qmodule


__all__ = ["QLayerNorm"]


@register_qmodule(torch.nn.LayerNorm)
class QLayerNorm(QModuleMixin, torch.nn.LayerNorm):
    @classmethod
    def from_module(cls, module, activations: Optional[torch.dtype] = None):
        if activations is None:
            return None
        qmodule = cls(
            module.normalized_shape,
            module.eps,
            module.elementwise_affine,
            module.bias is not None,
            activations=activations,
        )
        with torch.no_grad():
            qmodule.weight = torch.nn.Parameter(module.weight.clone().detach())
            if module.bias is not None:
                qmodule.bias = torch.nn.Parameter(module.bias.clone().detach())
        return qmodule.to(module.weight.device)

    def qforward(self, input: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.layer_norm(input, self.normalized_shape, self.weight, self.bias, self.eps)



import numbers
from functools import partial

import torch

from . import QTensor, dtype_info


__all__ = ["get_qtensor_op", "register_qtensor_op"]


_QTENSOR_OP_TABLE = {}


def register_qtensor_op(aten_ops):
    """
    Used for registering a new __torch_dispatch__ aten operation to QTensor

    The code to register a new operation looks like:

    @register_qtensor_op(list_of_ops)
    def foo(op, *args, **kwargs):
        <implementation>
    """

    def wrapper(op):
        for aten_op in aten_ops:
            _QTENSOR_OP_TABLE[aten_op] = partial(op, aten_op)

    return wrapper


def get_qtensor_op(aten_op):
    return _QTENSOR_OP_TABLE.get(aten_op, None)


def dequantized_op(func, *args, **kwargs):
    def dequantize(x):
        if isinstance(x, list):
            return [dequantize(y) for y in x]
        if isinstance(x, tuple):
            return (dequantize(y) for y in x)
        return x.dequantize() if isinstance(x, QTensor) else x

    dq_args = [dequantize(arg) for arg in args]
    dq_kwargs = {}
    for name, kwarg in kwargs.items():
        dq_kwargs[name] = dequantize(kwarg)
    return func(*dq_args, **dq_kwargs)


def ensure_qtensor_inputs(*args, per_tensor=True):
    for arg in args:
        if not isinstance(arg, QTensor):
            return False
        if per_tensor and arg.axis is not None:
            return False
    return True


def is_scalar(t):
    return isinstance(t, numbers.Number) or type(t) == torch.Tensor and len(t.shape) == 0


@register_qtensor_op([torch.ops.aten._to_copy])
def _to_copy(op, t, dtype=None, **kwargs):
    # For data, ignore dtype and use the inner type instead
    out_data = op(t._data, dtype=t.itype, **kwargs)
    # Apply the new dtype on the scale only
    out_scale = op(t._scale, dtype=dtype, **kwargs)
    return QTensor(out_data, out_scale)


@register_qtensor_op([torch.ops.aten.argmax])
def argmax(op, input, *args, **kwargs):
    if input.axis is not None:
        # If we have different scales we need to dequantize first
        return dequantized_op(op, input, *args, **kwargs)
    if input.itype.is_floating_point:
        # Argmax is not supported for float8
        return dequantized_op(op, input, *args, **kwargs)
    # We just return the argmax for the data
    return op(input._data, *args, **kwargs)


@register_qtensor_op([torch.ops.aten.detach])
def detach(op, t):
    # Detach both data and scale
    out_data = op(t._data)
    out_scale = op(t._scale)
    return QTensor(out_data, out_scale)


@register_qtensor_op([torch.ops.aten.add])
def add(op, input, other):
    # Only quantized tensors with identical scales cannot be added
    if isinstance(input, QTensor) and isinstance(other, QTensor) and torch.equal(input._scale, other._scale):
        # We need to perform the operation in int16 because it might overflow
        out_data = op(input._data.to(torch.int16), other._data.to(torch.int16))
        out_scale = input._scale
        return QTensor(out_data, out_scale)
    return dequantized_op(op, input, other)


@register_qtensor_op([torch.ops.aten.cat])
def cat(op, inputs, dim=0):
    if len(inputs) == 2:
        t1, t2 = inputs
        if isinstance(t1, QTensor) and isinstance(t2, QTensor) and torch.equal(t1._scale, t2._scale):
            if t1.itype.is_floating_point or t2.itype.is_floating_point:
                # Cat is not supported for float8
                return dequantized_op(op, inputs, dim)
            # Only quantized tensors with identical scales can be concatenated
            out_data = op([t1._data, t2._data], dim)
            return QTensor(out_data, t1._scale)
    return dequantized_op(op, inputs, dim)


@register_qtensor_op([torch.ops.aten.lt])
def lt(op, input, other):
    # Only quantized tensors with identical scales can be compared
    if isinstance(input, QTensor) and isinstance(other, QTensor) and torch.equal(input._scale, other._scale):
        return op(input._data, other._data)
    return dequantized_op(op, input, other)


@register_qtensor_op([torch.ops.aten.addmm])
def addmm(op, input, mat1, mat2, beta=1, alpha=1):
    if (
        alpha != 1
        or beta != 1
        or not ensure_qtensor_inputs(input, mat1, mat2, per_tensor=False)
        or mat1.axis is not None
    ):
        return dequantized_op(op, input, mat1, mat2, beta=beta, alpha=alpha)
    # Do the operation with int8 cast to float32
    out_data = op(
        input._data.to(torch.float32),
        mat1._data.to(torch.float32),
        mat2._data.to(torch.float32),
        beta=beta,
        alpha=alpha,
    )
    out_scale = mat1._scale * mat2._scale
    return QTensor(out_data.to(torch.int32), out_scale)


@register_qtensor_op([torch.ops.aten.clone])
def clone(op, t, memory_format=torch.preserve_format):
    out_data = op(t._data, memory_format=memory_format)
    out_scale = op(t._scale, memory_format=memory_format)
    return QTensor(out_data, out_scale)


@register_qtensor_op([torch.ops.aten.copy_])
def copy_(op, dest, src):
    dest._data = op(dest._data, src._data)
    dest._scale = op(dest._scale, src._scale)
    return dest


@register_qtensor_op([torch.ops.aten.div])
def div(op, input, other):
    if not is_scalar(other):
        return op(input.dequantize(), other)
    # We just divide the scale
    return QTensor(input._data, op(input._scale, other))


@register_qtensor_op([torch.ops.aten.dot])
def dot(op, input, other):
    if not ensure_qtensor_inputs(input, other):
        return dequantized_op(op, input, other)
    # Cast int8 data to float32 and do the operation
    out_data = op(input._data.to(torch.float32), other._data.to(torch.float32))
    out_scale = input._scale * other._scale
    return QTensor(out_data.to(torch.int32), out_scale)


@register_qtensor_op([torch.ops.aten.neg])
def neg(op, input, *args, **kwargs):
    if input.itype.is_floating_point:
        # Neg is not supported for float8
        return op(input.dequantize(), *args, **kwargs)
    out_data = op(input._data, *args, **kwargs)
    return QTensor(out_data, input._scale)


@register_qtensor_op(
    [
        torch.ops.aten.expand,
        torch.ops.aten.permute,
        torch.ops.aten.select,
        torch.ops.aten.slice,
        torch.ops.aten.unsqueeze,
    ]
)
def unary_type_agnostic_op(op, input, *args, **kwargs):
    if not ensure_qtensor_inputs(input):
        return op(input.dequantize(), *args, **kwargs)
    # These operations can be transparently applied on the underlying integer tensor,
    # without modifying the scale.
    out_data = op(input._data, *args, **kwargs)
    return QTensor(out_data, input._scale)


@register_qtensor_op([torch.ops.aten.is_same_size])
def is_same_size(op, input, other):
    a = input._data if isinstance(input, QTensor) else input
    b = other._data if isinstance(other, QTensor) else other
    return op(a, b)


@register_qtensor_op([torch.ops.aten.gelu, torch.ops.aten.masked_fill, torch.ops.aten.pow, torch.ops.aten.silu])
def unary_unsupported_op(op, input, *args, **kwargs):
    # Not supported: dequantize
    return op(input.dequantize(), *args, **kwargs)


@register_qtensor_op([torch.ops.aten.linear])
def linear(op, input, weight, bias=None):
    if (
        not isinstance(input, QTensor)
        or input.axis is not None
        or not isinstance(weight, QTensor)
        or (bias is not None and not isinstance(bias, QTensor))
    ):
        return dequantized_op(op, input, weight, bias=bias)
    # Cast int8 data to float32 and do the operation
    bias_data = bias._data.to(torch.float32) if bias is not None else None
    out_data = op(input._data.to(torch.float32), weight._data.to(torch.float32), bias_data)
    # The scalar input scale is broadcasted along all input dimensions
    input_scale = input._scale.view((1,) * input.ndim)
    # Weights are actually transposed inside the operation
    weight_scale = weight._scale.t()
    out_scale = input_scale * weight_scale
    return QTensor(out_data.to(torch.int32), out_scale)


@register_qtensor_op([torch.ops.aten.bmm])
def bmm(op, input, other):
    if not ensure_qtensor_inputs(input, other, per_tensor=False) or input.axis is not None:
        # Matric multiplication is only supported between a per-tensor QTensor and a QTensor
        return dequantized_op(op, input, other)
    # Cast int8 data to float32 and do the operation
    out_data = op(input._data.to(torch.float32), other._data.to(torch.float32))
    out_scale = input._scale * other._scale
    return QTensor(out_data.to(torch.int32), out_scale)


@register_qtensor_op([torch.ops.aten.mm])
def mm(op, input, other):
    if not ensure_qtensor_inputs(input, other, per_tensor=False) or input.axis is not None:
        # Matric multiplication is only supported between a per-tensor QTensor and a QTensor
        return dequantized_op(op, input, other)
    n, m = input.shape
    p = other.shape[-1]
    if (
        input.device.type == "cuda"
        and input.itype == torch.int8
        and other.itype == torch.int8
        and n > 16
        and n % 8 == 0
        and m % 8 == 0
        and p % 8 == 0
    ):
        # Use integer GEMM
        out_data = torch._int_mm(input._data, other._data)
    else:
        # Cast data to float32 and do the operation
        out_data = op(input._data.to(torch.float32), other._data.to(torch.float32))
    out_scale = input._scale * other._scale
    return QTensor(out_data.to(torch.int32), out_scale)


@register_qtensor_op([torch.ops.aten.mul])
def mul(op, input, other):
    # If one of the multiplicands is a scalar, just multiply the scale
    if is_scalar(input):
        return QTensor(other._data, input * other._scale)
    if is_scalar(other):
        return QTensor(input._data, other * input._scale)
    if not isinstance(input, QTensor) or not isinstance(other, QTensor):
        return dequantized_op(op, input, other)
    # Cast int8 data to int32 and do the operation
    out_data = op(input._data.to(torch.int32), other._data.to(torch.int32))
    out_scale = input._scale * other._scale
    return QTensor(out_data, out_scale)


@register_qtensor_op([torch.ops.aten.relu])
def relu(op, input):
    if input.itype.is_floating_point:
        # Relu is not supported for float8 types
        return dequantized_op(op, input)
    out_data = op(input._data)
    return QTensor(out_data, input._scale)


@register_qtensor_op([torch.ops.aten._softmax])
def _softmax(op, input, dim, half_to_float):
    # Softmax must be performed in float
    out_data = op(input.dequantize(), dim, half_to_float)
    # Since softmax is normalized, we know the optimal scale

    out_scale = torch.tensor(1 / dtype_info(input.itype).max, dtype=input.dtype).to(input.device)
    return QTensor.quantize(out_data, input.itype, out_scale)


@register_qtensor_op([torch.ops.aten.stack])
def stack(op, inputs, dim=0):
    if len(inputs) == 2:
        t1, t2 = inputs
        if isinstance(t1, QTensor) and isinstance(t2, QTensor) and torch.equal(t1._scale, t2._scale):
            # Only quantized tensors with identical scales can be stacked
            out_data = op([t1._data, t2._data], dim)
            return QTensor(out_data, t1._scale)
    return dequantized_op(inputs, dim)


@register_qtensor_op([torch.ops.aten.split])
def split(op, input, *args, **kwargs):
    out_datas = op(input._data, *args, **kwargs)
    return [QTensor(out_data, input._scale) for out_data in out_datas]


@register_qtensor_op([torch.ops.aten.transpose, torch.ops.aten.t])
def transpose(op, input, *args):
    out_data = op(input._data, *args)
    out_scale = input._scale
    if input.axis is not None:
        # We need to transpose also the scale
        out_scale = op(out_scale, *args)
    return QTensor(out_data, out_scale)


@register_qtensor_op([torch.ops.aten.view, torch.ops.aten._unsafe_view])
def view(op, input, *shape):
    out_data = op(input._data, *shape)
    if ensure_qtensor_inputs(input):
        return QTensor(out_data, input._scale)
    # We only support the simple case where the tensor is quantized along the
    # last axis that is not modified by the view
    if input.axis == input._scale.ndim - 1 and input._scale.shape[-1] == out_data.shape[-1]:
        out_scale_shape = (1,) * (out_data.ndim - 1) + (input._scale.shape[-1],)
        out_scale = input._scale.view(out_scale_shape)
        return QTensor(out_data, out_scale)
    return dequantized_op(op, input, *shape)


@register_qtensor_op([torch.ops.aten._softmax_backward_data])
def _softmax_backward_data(op, grad, output, dim, input_dtype):
    return op(grad, output.dequantize(), dim, input_dtype)


@register_qtensor_op([torch.ops.aten.threshold_backward])
def threshold_backward(op, grad, output, threshold):
    return op(grad, output.dequantize(), threshold)


@register_qtensor_op([torch.ops.aten.linear_backward])
def linear_backward(op, *args, **kwargs):
    return dequantized_op(op, *args, **kwargs)


@register_qtensor_op([torch.ops.aten.where])
def where(op, condition, input, other):
    if isinstance(condition, QTensor) or isinstance(other, QTensor):
        raise NotImplementedError
    float_data = op(condition, input.dequantize(), other)
    # We requantize with the input scale
    return QTensor.quantize(float_data, input.itype, input._scale)



from contextlib import contextmanager
from functools import partial

import torch
from torch.nn.modules.module import (
    register_module_forward_hook,
    register_module_forward_pre_hook,
)

from .nn import QModuleMixin
from .qtensor import QTensor, absmax_scale


__all__ = ["calibration"]


def updated_scale(scale, new_scale, momentum):
    if torch.all(scale == 1):
        return new_scale
    return momentum * scale + new_scale * (1.0 - momentum)


def calibrate_input(module: torch.nn.Module, input, momentum: float = 0.9):
    if isinstance(module, QModuleMixin) and module.activations is not None:
        input = input[0]
        if isinstance(input, QTensor):
            # Just adopt the maximum scale of the input
            module.input_scale = torch.max(input._scale)
        else:
            # Evaluate the best scale
            input_scale = absmax_scale(input, module.activations)
            module.input_scale = updated_scale(module.input_scale, input_scale, momentum)
        return input


def calibrate_output(
    module: torch.nn.Module,
    input,
    output,
    momentum: float = 0.9,
):
    if isinstance(module, (QModuleMixin)) and module.activations is not None:
        # Reevaluate raw module output
        qoutput = module.qforward(input[0])
        if isinstance(qoutput, QTensor):
            qoutput = qoutput.dequantize()
        # Evaluate the optimal scale per-tensor and update output scale
        output_scale = absmax_scale(qoutput, module.activations, axis=None)
        module.output_scale = updated_scale(module.output_scale, output_scale, momentum)
        # Reevaluate output with the correct output scale
        return module.forward(input[0])


@contextmanager
def calibration(momentum: float = 0.9):
    """A context to evaluate the quantized modules input and output scales.

    Scales are evaluated per-batch using the absmax algorithm and aggregated using a
    momentum.
    Input and output scales are always evaluated per-tensor.

    Args:
        momentum (`float`): the momentum to use when updating scales.
    """
    try:
        pre_handle = register_module_forward_pre_hook(partial(calibrate_input, momentum=momentum))
        post_handle = register_module_forward_hook(partial(calibrate_output, momentum=momentum))
        yield
    finally:
        pre_handle.remove()
        post_handle.remove()



from functools import partial

import torch

from .core import QTensor


__all__ = ["get_qtensor_func", "register_qtensor_func"]


_QTENSOR_FUNC_TABLE = {}


def register_qtensor_func(funcs):
    """
    Used for registering a new __torch_dispatch__ function to QTensor

    The code to register a new function looks like:

    @register_qtensor_func(list_of_funcs)
    def foo(func, *args, **kwargs):
        <implementation>
    """

    def wrapper(qfunc):
        for func in funcs:
            _QTENSOR_FUNC_TABLE[func] = partial(qfunc, func)

    return wrapper


def get_qtensor_func(func):
    return _QTENSOR_FUNC_TABLE.get(func, None)


def dequantize(*args):
    return [arg.dequantize() if isinstance(arg, QTensor) else arg for arg in args]


@register_qtensor_func([torch.nn.functional.log_softmax, torch.topk, torch.nn.functional.layer_norm])
def unary_unsupported_op(func, t, *args, **kwargs):
    return func(t.dequantize(), *args, **kwargs)


@register_qtensor_func([torch.nn.functional.cross_entropy, torch.nn.functional.cosine_similarity])
def plurary_unsupported_op(func, *args, **kwargs):
    return func(*dequantize(*args), **kwargs)
